"""Strategies for generating PyTorch Modules."""

from __future__ import annotations

from typing import TypeVar

import torch
from hypothesis import strategies as st
from torch import nn

from hypothesis_torch import utils

__all__ = [
    "same_shape_activation_strategy",
    "linear_network_strategy",
]

T = TypeVar("T")

SENSIBLE_FLOATS = st.floats(
    min_value=-10,
    max_value=10,
    allow_nan=False,
    allow_infinity=False,
    allow_subnormal=False,
)
SENSIBLE_POSITIVE_FLOATS = st.floats(
    min_value=0,
    max_value=10,
    allow_nan=False,
    allow_infinity=False,
    allow_subnormal=False,
)
POSITIVE_INTS = st.integers(min_value=1)


@st.composite
def signature_to_strategy(draw: st.DrawFn, constructor: type[T], *args, **kwargs) -> T:
    """Strategy for generating instances of a class by drawing values for its constructor.

    Args:
    ----
        draw: The draw function provided by `hypothesis`.
        constructor: The class to generate an instance of.
        args: Positional arguments to pass to the constructor. If an argument is a strategy, it will be drawn from.
        kwargs: Keyword arguments to pass to the constructor. If a keyword argument is a strategy, it will be drawn
            from.

    Returns:
    -------
        An instance of the class.

    """
    args_drawn = [draw(strategy) for strategy in args]
    kwargs_drawn = {k: draw(strategy) for k, strategy in kwargs.items()}
    return constructor(*args_drawn, **kwargs_drawn)


@st.composite
def hard_tanh_strategy(draw: st.DrawFn) -> nn.Hardtanh:
    """Strategy for generating instances of `nn.Hardtanh` by drawing values for its constructor.

    Args:
    ----
        draw: The draw function provided by `hypothesis`.

    Returns:
    -------
        An instance of `nn.Hardtanh`.

    """
    min_value = draw(SENSIBLE_FLOATS)
    max_value = draw(SENSIBLE_FLOATS.filter(lambda x: x > min_value))
    inplace = draw(st.booleans())
    return nn.Hardtanh(min_value, max_value, inplace)


def same_shape_activation_strategy() -> st.SearchStrategy[nn.Module]:
    """Strategy for generating activation functions that have the same shape input and output shape.

    Returns:
        A strategy for generating activation functions that have the same shape input and output shape.
    """
    return st.one_of(
        signature_to_strategy(nn.Identity),
        signature_to_strategy(nn.ELU, alpha=SENSIBLE_FLOATS, inplace=st.booleans()),
        signature_to_strategy(nn.Hardshrink, lambd=SENSIBLE_FLOATS),
        signature_to_strategy(nn.Hardsigmoid, inplace=st.booleans()),
        hard_tanh_strategy(),
        signature_to_strategy(nn.Hardswish, inplace=st.booleans()),
        signature_to_strategy(nn.LeakyReLU, negative_slope=SENSIBLE_FLOATS, inplace=st.booleans()),
        signature_to_strategy(nn.LogSigmoid),
        # TODO: nn.MultiheadAttention, although in the `Non-linear activations` section, does not have the same shape
        #  inside and outside
        # TODO: nn.PReLU(num_parameters=1, init=0.25, device=None, dtype=None)
        # TODO: PReLU might depend on the input shape
        # TODO: num_parameters (int) – number of a to learn. Although it takes an int as input, there is only two
        #  values are legitimate: 1, or the number of channels at input. Default: 1
        signature_to_strategy(nn.PReLU, num_parameters=st.just(1), init=SENSIBLE_FLOATS),
        signature_to_strategy(nn.ReLU, inplace=st.booleans()),
        signature_to_strategy(nn.ReLU6, inplace=st.booleans()),
        signature_to_strategy(
            nn.RReLU,
            lower=SENSIBLE_FLOATS,
            upper=SENSIBLE_FLOATS,
            inplace=st.booleans(),
        ),
        signature_to_strategy(nn.SELU, inplace=st.booleans()),
        signature_to_strategy(nn.CELU, alpha=SENSIBLE_FLOATS, inplace=st.booleans()),
        signature_to_strategy(nn.GELU, approximate=st.sampled_from(["none", "tanh"])),
        signature_to_strategy(nn.Sigmoid),
        signature_to_strategy(nn.SiLU, inplace=st.booleans()),
        signature_to_strategy(nn.Mish, inplace=st.booleans()),
        signature_to_strategy(nn.Softplus, beta=SENSIBLE_FLOATS, threshold=POSITIVE_INTS),
        signature_to_strategy(nn.Softshrink, lambd=SENSIBLE_POSITIVE_FLOATS),
        signature_to_strategy(nn.Softsign),
        signature_to_strategy(nn.Tanh),
        signature_to_strategy(nn.Tanhshrink),
        signature_to_strategy(
            nn.Threshold,
            threshold=SENSIBLE_FLOATS,
            value=SENSIBLE_FLOATS,
            inplace=st.booleans(),
        ),
        # TODO: nn.GLU depends on the input shape
    )


@st.composite
def linear_network_strategy(
    draw: st.DrawFn,
    input_shape: tuple[int, ...] | torch.Size | st.SearchStrategy[tuple[int, ...]] | st.SearchStrategy[torch.Size],
    output_shape: tuple[int, ...] | torch.Size | st.SearchStrategy[tuple[int, ...]] | st.SearchStrategy[torch.Size],
    activation_layer: nn.Module | st.SearchStrategy[nn.Module],
    hidden_layer_size: int | st.SearchStrategy[int],
    depth: int | st.SearchStrategy[int],
    device: torch.device | st.SearchStrategy[torch.device],
) -> nn.Module:
    """Strategy for generating random Torch sequential networks of linear layers with activation functions.

    Args:
        draw: The draw function provided by `hypothesis`.
        input_shape: The shape of the input tensor. If a strategy is provided, it will be drawn from.
        output_shape: The shape of the output tensor. If a strategy is provided, it will be drawn from.
        activation_layer: Activation layer to use. If a strategy is provided, it will be drawn from.
        hidden_layer_size: The size of the hidden layers. If a strategy is provided, it will be drawn from.
        depth: The maximum depth of the network. If a strategy is provided, it will be drawn from.
        device: The device on which to place the network. If a strategy is provided, it will be drawn from.

    Returns:
        A strategy for generating random linear networks.

    """
    if isinstance(input_shape, st.SearchStrategy):
        input_shape = draw(input_shape)
    if len(input_shape) not in {1, 2}:
        raise ValueError(f"input_shape must be a 1D or 2D tuple, but got {input_shape}")
    *_, input_size = input_shape

    if isinstance(output_shape, st.SearchStrategy):
        output_shape = draw(output_shape)
    if len(output_shape) not in {1, 2}:
        raise ValueError(f"output_shape must be a 1D or 2D tuple, but got {output_shape}")
    *_, output_size = output_shape

    if isinstance(device, st.SearchStrategy):
        device = draw(device)

    if isinstance(activation_layer, nn.Module):
        activation_layer_strategy = st.just(activation_layer)
    else:
        activation_layer_strategy = activation_layer

    if not isinstance(hidden_layer_size, st.SearchStrategy):
        hidden_layer_size = st.just(hidden_layer_size)
    if isinstance(depth, st.SearchStrategy):
        depth = draw(depth)

    with device:
        interior_layer_sizes = draw(
            st.lists(
                hidden_layer_size,
                min_size=depth,
                max_size=depth,
            ),
        )
        layer_sizes = [input_size, *interior_layer_sizes, output_size]
        layers: list[nn.Module] = [nn.Linear(a, b) for a, b in utils.pairwise(layer_sizes)]
        activations: list[nn.Module] = draw(
            st.lists(activation_layer_strategy, min_size=len(layers), max_size=len(layers)),
        )
        return nn.Sequential(*utils.alternate(layers, activations))


def convolution_output_shape(
    input_shape: tuple[int, ...] | torch.Tensor,
    kernel_size: tuple[int, ...] | torch.Tensor,
    stride: tuple[int, ...] | torch.Tensor,
    padding: tuple[int, ...] | torch.Tensor,
    dilation: tuple[int, ...] | torch.Tensor,
) -> tuple[int, ...]:
    """Calculate the output shape of a convolutional layer."""
    input_shape = torch.tensor(input_shape, dtype=torch.uint8)
    kernel_size = torch.tensor(kernel_size, dtype=torch.uint8)
    stride = torch.tensor(stride, dtype=torch.uint8)
    padding = torch.tensor(padding, dtype=torch.uint8)
    dilation = torch.tensor(dilation, dtype=torch.uint8)
    return tuple(torch.floor((input_shape + 2 * padding - dilation * (kernel_size - 1) - 1) / stride + 1).tolist())



"""Root imports for hypothesis-torch."""

__version__ = "0.1.8"
import importlib.util

from hypothesis_torch.device import device_strategy
from hypothesis_torch.dtype import dtype_strategy
from hypothesis_torch.module import linear_network_strategy, same_shape_activation_strategy
from hypothesis_torch.tensor import tensor_strategy

if importlib.util.find_spec("transformers") is not None:
    # Import Hugging Face strategies if transformers is installed
    from hypothesis_torch.huggingface import transformer_strategy



"""Strategies for generating PyTorch tensors."""

from __future__ import annotations

from collections.abc import Mapping
from typing import Any

import hypothesis.extra.numpy as numpy_st
import torch
from hypothesis import strategies as st

from hypothesis_torch import dtype as dtype_module


@st.composite
def tensor_strategy(
    draw: st.DrawFn,
    dtype: torch.dtype | st.SearchStrategy[torch.dtype],
    shape: int | st.SearchStrategy[int] | tuple[int, ...] | st.SearchStrategy[tuple[int, ...]],
    *,
    elements: st.SearchStrategy[Any] | Mapping[str, Any] | None = None,
    fill: st.SearchStrategy[Any] | None = None,
    unique: bool | st.SearchStrategy[bool] = False,
    device: torch.device | st.SearchStrategy[torch.device] | None = None,
) -> torch.Tensor:
    """A strategy for generating PyTorch tensors.

    Args:
        draw: The draw function provided by `hypothesis`.
        dtype: Any PyTorch dtype or a strategy for generating PyTorch dtypes.
        shape: The shape of the tensor. Can be an integer >= 0, a tuple of such integers, or a strategy for generating
            such values.
        elements: A strategy for generating elements of the tensor. If `None`, a suitable strategy is inferred from
            the dtype. Note that this may give any legal value (including NaNs and infinities for floats).
        fill: A strategy for generating a single background value for the tensor. If None, a suitable default will
            be inferred based on the other arguments. If set to `~hypothesis.strategies.nothing` then filling
            behaviour will be disabled entirely and every element will be generated independently.
        unique: Whether the tensor's elements should all be distinct from one another. Note that multiple NaN values
            may still be allowed.
        device: The device on which to place the tensor. If None, the default device is used.

    Returns:
        A strategy for generating PyTorch tensors.

    """
    # We will pre-sample the dtype so that we can cast it to a concrete numpy dtype
    if isinstance(dtype, st.SearchStrategy):
        dtype = draw(dtype)
    numpy_dtype = dtype_module.numpy_dtype_map[dtype]

    if isinstance(unique, st.SearchStrategy):
        unique = draw(unique)

    ndarray_strategy = numpy_st.arrays(numpy_dtype, shape, elements=elements, fill=fill, unique=unique)
    tensor = draw(ndarray_strategy.map(torch.from_numpy))

    # We will pre-sample the device so that we can cast it to a concrete torch device
    if isinstance(device, st.SearchStrategy):
        device = draw(device)

    return tensor.to(dtype=dtype, device=device)  # Final casting to the desired dtype and device


st.register_type_strategy(
    torch.Tensor,
    tensor_strategy(
        dtype=dtype_module.dtype_strategy(),
        shape=st.integers(min_value=0, max_value=10),  # Discourage large tensors
    ),
)



"""Strategies for generating Hugging Face transformers models."""

from __future__ import annotations

import inspect
from typing import TypeVar

import hypothesis.strategies as st
import torch
import transformers

from hypothesis_torch import inspection_util

T = TypeVar("T")
TransformerType = TypeVar("TransformerType", bound=transformers.PreTrainedModel)

POSITIVE_INTS = st.integers(min_value=1)
FLOATS_BETWEEN_ZERO_AND_ONE = st.floats(
    min_value=0,
    max_value=1,
    allow_subnormal=False,
    allow_nan=False,
    allow_infinity=False,
)
FLOATS_GREATER_THAN_ZERO = st.floats(min_value=0, allow_subnormal=False, allow_nan=False, allow_infinity=False)
FLOATS_GREATER_THAN_ONE = st.floats(min_value=1, allow_subnormal=False, allow_nan=False, allow_infinity=False)


TRANSFORMER_CONFIG_KWARG_STRATEGIES = {
    "attention_bias": st.booleans(),
    "attention_dropout": FLOATS_BETWEEN_ZERO_AND_ONE,
    "bos_token_id": POSITIVE_INTS,
    "eos_token_id": POSITIVE_INTS,
    "hidden_act": st.sampled_from(["gelu", "relu", "silu", "gelu_new", "tanh"]),
    "hidden_size": POSITIVE_INTS,
    "initializer_range": FLOATS_GREATER_THAN_ZERO,
    "intermediate_size": POSITIVE_INTS,
    "max_position_embeddings": POSITIVE_INTS,
    "num_attention_heads": POSITIVE_INTS,
    "num_hidden_layers": POSITIVE_INTS,
    "rms_norm_eps": FLOATS_GREATER_THAN_ZERO,
    "pretrained_tp": POSITIVE_INTS,
    "rope_theta": FLOATS_GREATER_THAN_ZERO,
}


@st.composite
def build_from_cls_init(draw: st.DrawFn, cls: type[T], **kwargs) -> T:
    """Strategy for generating instances of a class by drawing values for its constructor.

    Args:
        draw: The draw function provided by `hypothesis`.
        cls: The class to generate an instance of.
        kwargs: Keyword arguments to pass to the constructor. If a keyword argument is a strategy, it will be drawn
            from.

    Returns:
        An instance of the class.

    """
    sig = inspection_util.infer_signature_annotations(cls.__init__)
    for param in sig.parameters.values():
        if param.name in kwargs and isinstance(kwargs[param.name], st.SearchStrategy):
            kwargs[param.name] = draw(kwargs[param.name])
        if param.annotation is inspect.Parameter.empty:
            continue
        if param.name in TRANSFORMER_CONFIG_KWARG_STRATEGIES:
            kwargs[param.name] = draw(TRANSFORMER_CONFIG_KWARG_STRATEGIES[param.name])
        else:
            kwargs[param.name] = draw(st.from_type(param.annotation))
    kwargs.pop("self", None)  # Remove self if a type was inferred
    return cls(**kwargs)


@st.composite
def transformer_strategy(
    draw: st.DrawFn,
    cls: type[TransformerType] | st.SearchStrategy[type[TransformerType]],
    *,
    instantiate_weights: bool | st.SearchStrategy[bool] = True,
    **kwargs,
) -> TransformerType:
    """Strategy for generating Hugging Face transformers.

    Args:
        draw: The draw function provided by `hypothesis`.
        cls: The transformer class to generate.
        instantiate_weights: Whether to instantiate the weights of the transformer. If False, the transformer will be
            instantiated on the meta device. This is useful for testing uses of transformers models that do not require
            a forward pass.
        kwargs: Keyword arguments to pass to the transformer constructor. If a keyword argument is a strategy, it will
            be drawn from.

    Returns:
        A strategy for generating Hugging Face transformers.
    """
    if isinstance(cls, st.SearchStrategy):
        cls = draw(cls)
    assert issubclass(cls, transformers.PreTrainedModel)
    config = draw(build_from_cls_init(cls.config_class, **kwargs))

    if isinstance(instantiate_weights, st.SearchStrategy):
        instantiate_weights = draw(instantiate_weights)

    if instantiate_weights:
        return cls(config)
    with torch.device("meta"):
        return cls(config)



"""Strategies for generating PyTorch dtypes."""

from __future__ import annotations

from collections.abc import Sequence

import numpy as np
import numpy.typing as npt
import torch
from hypothesis import strategies as st

SIGNED_INT_DTYPES = [torch.int8, torch.int16, torch.int32, torch.int64]
"""All signed integer dtypes supported by PyTorch."""
UNSIGNED_INT_DTYPES = [torch.uint8]
"""All unsigned integer dtypes supported by PyTorch."""
INT_DTYPES = SIGNED_INT_DTYPES + UNSIGNED_INT_DTYPES
"""All integer (both signed and unsigned) dtypes supported by PyTorch."""

FLOAT_DTYPES = [torch.float16, torch.float32, torch.float64]
"""All floating point dtypes supported by PyTorch."""
BFLOAT_DTYPES = [torch.bfloat16]
"""All brain-float dtypes supported by PyTorch."""
COMPLEX_DTYPES = [torch.complex64, torch.complex128]
"""All complex dtypes supported by PyTorch."""
BOOL_DTYPES = [torch.bool]
"""All boolean dtypes supported by PyTorch."""

ALL_DTYPES = INT_DTYPES + FLOAT_DTYPES + BFLOAT_DTYPES + COMPLEX_DTYPES + BOOL_DTYPES
"""All dtypes supported by PyTorch."""

numpy_dtype_map: dict[torch.dtype, npt.DTypeLike] = {
    torch.int8: np.int8,
    torch.int16: np.int16,
    torch.int32: np.int32,
    torch.int64: np.int64,
    torch.uint8: np.uint8,
    torch.float16: np.float16,
    torch.float32: np.float32,
    torch.float64: np.float64,
    torch.bfloat16: np.float32,  # bfloat16 is not supported by numpy
    torch.complex64: np.complex64,
    torch.complex128: np.complex128,
    torch.bool: np.bool_,
}
"""A mapping from torch dtypes to numpy dtypes. Useful for generating tensors of arbitrary dtypes from the builtin
numpy strategies."""

assert set(numpy_dtype_map.keys()) == set(ALL_DTYPES)


@st.composite
def dtype_strategy(draw: st.DrawFn, *, dtypes: Sequence[torch.dtype] | None = None) -> torch.dtype:
    """Strategy for generating torch dtypes.

    Args:
        draw: The draw function provided by `hypothesis`. If `None`, all dtypes are sampled.
        dtypes: A strategy for generating elements of the dtype.

    Returns:
        A strategy for generating torch dtypes.

    """
    if dtypes is None:
        dtypes = ALL_DTYPES
    return draw(st.sampled_from(dtypes))


st.register_type_strategy(torch.dtype, dtype_strategy())
